/*
 * Pluginsystem
 * Copyright (C) Thasso Griebel
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package org.java.plugin;

import com.google.inject.*;
import com.google.inject.name.Named;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.java.plugin.PluginManager.PluginLocation;
import org.java.plugin.extension.InstanceFactory;
import org.java.plugin.extension.InternalPluginLocation;
import org.java.plugin.extension.PluginLifecycleListener;
import org.java.plugin.extension.PluginsCollector;
import org.java.plugin.registry.IntegrityCheckReport;
import org.java.plugin.registry.PluginDescriptor;
import org.java.plugin.registry.PluginRegistry;
import org.java.plugin.util.ExtendedProperties;

import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;
import java.util.Properties;

/**
 * The plugin Controller provides methods to initialize the plugin framework, load
 * plugin location and access plugins
 */
@Singleton
public class PluginController {
    private static final Log logger = LogFactory.getLog(PluginController.class);

    /**
	 * The JPF Plugin Manager
	 */
	private org.java.plugin.PluginManager pluginManager;

    /**
	 * The global configuration
	 */
	private Properties configuration;

    /**
     * The general plugins classloader
     */
    private ClassLoader pluginsClassLoader;
    /**
     * The plugins collector
     */
    private PluginsCollector collector;
    /**
     * the object factory
     */
    private ObjectFactory objectFactory;
    /**
     * Store lifecycel listeners here
     */
    private List<PluginLifecycleListener> listenerList;

    @Inject
    public PluginController(@Named("PluginsClassLoader")ClassLoader pluginsClassLoader, ObjectFactory objectFactory){
        super();
        this.pluginsClassLoader = pluginsClassLoader;
        this.objectFactory = objectFactory;
        this.listenerList = new ArrayList<PluginLifecycleListener>();
    }

    /**
     * Adds a {@link PluginLifecycleListener}. Does
     * nothing when listener was added before
     *
     * @param listener the listener
     * @see PluginLifecycleListener
     */
    public synchronized void addPluginLifecycleListener(PluginLifecycleListener listener){
        if(listenerList.contains(listener)) return; // allready added
        listenerList.add(listener);
        if(pluginManager != null){
            pluginManager.registerListener(listener);
        }
    }

    /**
     * Removed {@link PluginLifecycleListener}. Does nothing when
     * listener was not attached.
     *
     * @param listener the listener
     * @see PluginLifecycleListener
     */
    public synchronized void removePluginLifecycleListener(PluginLifecycleListener listener){
        listenerList.remove(listener);
        if(pluginManager != null){
            try{
                pluginManager.unregisterListener(listener);
            }catch(IllegalArgumentException ignored){} // thrown when listener wasn't added before
        }

    }

    /*
	 * Basic framework initialization
     *  - loadCofiguration
	 */
	public void initialize(Properties config) {
		loadConfiguration(config);
		// collect plugins
		collector = new PluginsCollector(getConfiguration());


		// set Thread class loader to AllPluginsclassloader
		// this is used by the StandardPluginManager while
		// queried for plugin classloaders. It registers the
		// classloaders and - at the same time, the AllPluginsClassloader
		// is used as parent loader (they use Thread.getContextclassloader()),
		// which lets the pluginsclassloader itself use the AllPluginsclassloader
		// to find classes and resources
		Thread.currentThread().setContextClassLoader(pluginsClassLoader);

		// initialize PluginController
        // we need a PathResolver
		configureObjectFactory();
		PathResolver pathResolver = objectFactory.createPathResolver();
		PluginRegistry pluginRegistry = objectFactory.createRegistry();
		pluginManager = objectFactory.createManager(pluginRegistry, pathResolver);
        // append listeners
        if(listenerList.size() > 0){
            for (PluginLifecycleListener listener : listenerList) {
                pluginManager.registerListener(listener);
            }
        }

        /*
        Register private local plugin
         */
        try {
            URL internalPlugin = getClass().getResource("/META-INF/pluginsystem.xml");            
            registerPlugins(internalPlugin);
        } catch (JpfException e) {
            logger.fatal("Error while registering internal PluginSystem plugin!", e);
        }
        
        try {
			// register other plugins
			PluginLocation[] locations = collector.collectPluginLocations();
			pluginManager.publishPlugins(locations);
		} catch (JpfException e) {
			logger.fatal("Error while starting plugin manager!", e);
		}

		// do integrity check ?
		String icheck = getConfiguration().getProperty("Application.integrityCheck", "off");
		if (!icheck.equalsIgnoreCase("off")) {
			integrityCheck();
		}

	}

    /**
	 * Translates Application configuration to JPF ExtendedProperties before it
	 * configures and returns a new {@link ObjectFactory}. 
	 */
	protected void configureObjectFactory() {

		ExtendedProperties jpfConf = new ExtendedProperties(getConfiguration());
		// plugins collector, map the sorther jpf.plugindir
		jpfConf.put("org.java.plugin.boot.pluginsRepositories", jpfConf.getProperty("jpf.plugindir", ""));
        
		// shading path resolver and target directory
		jpfConf.put("org.java.plugin.standard.ShadingPathResolver.shadowFolder",
            System.getProperty("java.io.tmpdir") + File.separator + "plugin-cache");
        objectFactory.configure(jpfConf);
	}


    /**
     * Access the global classloader
     *
     * @return classloader in global scope
     */
    public ClassLoader getPluginsClassLoader() {
        return pluginsClassLoader;
    }

    /**
     * Access the Plugins Classloader
     *
     * @return classloader for given plugin descriptor
     */
    public ClassLoader getPluginClassLoader(PluginDescriptor descriptor) {
        return getJPFManager().getPluginClassLoader(descriptor);
    }


    /**
     * @return jpfManager 
     */
    public org.java.plugin.PluginManager getJPFManager() {
        return pluginManager;
    }


    /**
     * Register a new plugin folder. This adds the given directory
     * as plugin repository and restarts scanning.
     *
     * @param location the new location
     * @return true if successfully registered
     * @throws JpfException when an error occures while scanning for plugins
     */
    public boolean registerLocation(File... location) throws JpfException {
        // cache old locations
        for (File file : location) {
            if(!getPluginsCollector().addRepository(file)){
                //return false;
                // do not return here, just report a warning
                logger.warn("Plugin Repository " + file + " could not be added (meybe it already exists ?)");
            }
        }
        // we added all repository successfully
        // now rescanfor plugins
        logger.debug("Added repositories - rescanning locations");
        PluginLocation[] locations = collector.collectPluginLocations();
        pluginManager.publishPlugins(locations);
        return true;
    }



    /**
     * Register custom plugins.
     *
     * @param locations the locations to register
     * @throws JpfException in case registration fails
     */
    void registerPlugins(PluginLocation... locations) throws JpfException {
        /*
        Implements #6 manual plugin registration
         */        
        getJPFManager().publishPlugins(locations);
    }
    
    /**
     * Register custom plugins. The given URLs must point to valid plugin
     * definition XML files. 
     *
     * @param pluginURLs the locations to register
     * @throws JpfException in case registration fails
     */
    public void registerPlugins(URL... pluginURLs) throws JpfException {
        PluginLocation[] locs = new PluginLocation[pluginURLs.length];
        for (int i = 0; i < locs.length; i++) {
            if(pluginURLs[i] == null){
                throw new NullPointerException("Null Plugin url is not allowed !");
            }
            locs[i] = new InternalPluginLocation(pluginURLs[i]);            
        }
        registerPlugins(locs);
    }

    /**
     * Access the plugins collector
     *
     * @return collector used to find plugin locations
     */
    PluginsCollector getPluginsCollector(){
        return collector;
    }

    

    public void shutdown() {
		// bring down pluginManager
		if (pluginManager != null) {
			pluginManager.shutdown();
		}
		// shut down logging
		LogFactory.releaseAll();
	}

    /**
	 * Load boot configuration.<br>
	 * The configuration defaults are hardcoded here and can easily be overwritten using a
	 * resource bundle. Also, overriding properties can be done using system properties or via command
	 * line configuration. <br>
	 * The load order is defaults->system properties->cmd line options
     *
     * @param config runtime configuration
	 */
	protected void loadConfiguration(Properties config) {
        Properties jpfConf = new Properties();
        try {
            jpfConf.load(getClass().getResource("/META-INF/jpf.properties").openStream());
        } catch (IOException e) {
            logger.error("Error while reading jpf properties!", e);
        }

        // override from config
        if(config != null){
            for (Object key : config.keySet()) {
                if(config.get(key) != null){
                    jpfConf.put(key, config.get(key));
                }
            }
        }

		// override from system properties
		// get the keys
		Object[] keys = new Object[jpfConf.keySet().size()];
		keys = jpfConf.keySet().toArray(keys);
		for (Object key : keys) {
			String value = System.getProperty(key.toString());
			if (value != null) {
				// override property
				jpfConf.put(key, value);
			}
		}
		configuration = jpfConf;
	}


    /**
	 * Returns the current configuration
	 *
	 * @return configuration of this plugin manager
	 */
	public Properties getConfiguration() {
        loadConfiguration(configuration);
		return configuration;
	}

    // ///////// HELPER
	//
	protected void integrityCheck() {
		// Check plug-in's integrity.
		logger.debug("checking plug-ins set integrity");
		String check = getConfiguration().getProperty("jpf.integrityCheck", "light");

		// I dont understand ? we nulling the path resolver when check is set to light ?
		IntegrityCheckReport integrityCheckReport = pluginManager.getRegistry()
				.checkIntegrity(check.equals("light") ? null : pluginManager.getPathResolver());
		logger.info("integrity check done: errors - " //$NON-NLS-1$
				+ integrityCheckReport.countErrors() + ", warnings - " //$NON-NLS-1$
				+ integrityCheckReport.countWarnings());
		if (integrityCheckReport.countErrors() != 0) {
			// Something wrong in plug-ins set.
			logger.info(integrityCheckReport2str(integrityCheckReport));
			// TODO: Handle error -> log file + gui dialog ?
		} else if (logger.isDebugEnabled()
				&& ((integrityCheckReport.countErrors() > 0) || (integrityCheckReport
						.countWarnings() > 0))) {
			logger.debug(integrityCheckReport2str(integrityCheckReport));
		}
	}

    protected String integrityCheckReport2str(final IntegrityCheckReport report) {
		StringBuffer buf = new StringBuffer();
		buf.append("integrity check report:\r\n"); //$NON-NLS-1$
		buf.append("-------------- REPORT BEGIN -----------------\r\n"); //$NON-NLS-1$
        for (Object o : report.getItems()) {
            IntegrityCheckReport.ReportItem item = (IntegrityCheckReport.ReportItem) o;
            buf.append("\tseverity=").append(item.getSeverity()) //$NON-NLS-1$
                    .append("; code=").append(item.getCode()) //$NON-NLS-1$
                    .append("; message=").append(item.getMessage()) //$NON-NLS-1$
                    .append("; source=").append(item.getSource()) //$NON-NLS-1$
                    .append("\r\n"); //$NON-NLS-1$
        }
		buf.append("-------------- REPORT END -----------------"); //$NON-NLS-1$
		return buf.toString();
	}

    ////////////end helper

    /**
     * Looks for plug-in with given ID and activates it if it is not activated
     * yet. Note that this method will never return <code>null</code>.
     *
     * @param uniqueID unique plugin id
     * @return plugin or null
     * @throws PluginLifecycleException if plugin could not be resolved
     */
    public Plugin getPlugin(String uniqueID) throws PluginLifecycleException {
        return pluginManager.getPlugin(uniqueID);
    }

    /**
     * Activate plugin. The Plugin becomes activated if it is not already
     * active or in activating stage.
     *
     * @param plugin the plugin to be started
     * @throws PluginLifecycleException if an error occures while the plugin is started
     */
    public void activatePlugin(Plugin plugin) throws PluginLifecycleException {
        activatePlugin(plugin.getDescriptor());
    }
    /**
     * Activate plugin. The Plugin becomes activated if it is not already
     * active or in activating stage.
     *
     * @param descriptor the plugin descriptor
     * @throws PluginLifecycleException if an error occures while the plugin is started
     */
    public void activatePlugin(PluginDescriptor descriptor) throws PluginLifecycleException {
        if(!getJPFManager().isPluginActivated(descriptor) && !getJPFManager().isPluginActivating(descriptor))
            getJPFManager().activatePlugin(descriptor.getId());
    }
    /**
     * Activate plugin. The Plugin becomes activated if it is not already
     * active or in activating stage.
     *
     * @param id the plugin to be started
     * @throws PluginLifecycleException if an error occures while the plugin is started
     */
    public void activatePlugin(String id) throws PluginLifecycleException {
        activatePlugin(getPlugin(id).getDescriptor());

    }

    /**
     * Returns a list of all currently available PluginDescriptors
     * 
     * @return list of PluginDescriptors
     * @see org.java.plugin.registry.PluginDescriptor
     */
    public List<PluginDescriptor> getPluginDescriptions() {
        return new ArrayList<PluginDescriptor>(pluginManager.getRegistry().getPluginDescriptors());
    }

    ////////// static access
    //
    //
    /**
     * Create a new PluginController with default configuration
     *
     * @return pluginController with default configuration
     */
    public static PluginController create(){
        return create(new Properties());
    }

    /**
     * Create a new PluginController  with a custom runtime configuration
     *
     * @param config the runtime configuration
     * @return pluginController with custom configuration
     */
    public static PluginController create(Properties config) {
        return create(config, new ArrayList<Module>());
    }

    /**
     * Create a new PluginController with a custom runtime configuration and a list
     * of Guice Modules used to configure the main {@link com.google.inject.Injector}.
     *
     * @param config the runtime configuration
     * @param modules a list of modules
     * @return pluginController with custom configuration
     */
    public static PluginController create(Properties config, List<? extends Module> modules ) {
        if(config == null){
            config = new Properties();
        }
        PluginControllerModule pluginControllerModule = new PluginControllerModule();
        ArrayList<Module> moduleList = new ArrayList<Module>();
        moduleList.add(pluginControllerModule);
        if(modules != null)
            moduleList.addAll(modules);
        Injector injector = Guice.createInjector(moduleList);
        PluginController controller = injector.getInstance(PluginController.class);
        controller.initialize(config);
        // prepare InstanceFactory
        injector.getInstance(InstanceFactory.class).initiazlieFromController();
        return controller;
    }
    //
    //
    ///////// end static access
    
}
