/*
 * Pluginsystem
 * Copyright (C) Thasso Griebel
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package org.java.plugin.extension;

import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.*;

import org.java.plugin.PluginClassLoader;
import com.google.inject.Inject;

/*
    - creating defensive copies to avoid concrrent modification exceptions
 */

/**
 * Helper class to collect all instances of {@link PluginClassLoader}s in use.
 * We can use this loader to find classes and resources of activated plugins
 * even without an explicit import statement.
 * <p>
 * This is mostly useful for scripting purposes and you have to keep in mind that
 * plugins still have to be activated, otherwise their classloader is either not
 * registered or it will throw a {@link ClassNotFoundException}.
 *
 * 
 * @author Thasso Griebel
 *
 */
public class AllPluginsClassLoader extends ClassLoader{
    /**
     * Singelton instance
     */
    private static AllPluginsClassLoader instance;

    /**
     * Store all registered classloaders
     */
    private Vector<ClassLoader> loader = new Vector<ClassLoader>();
    /**
     * Create a new instance.
     * 
     * @param parent parent loader
     */
    @Inject
    private AllPluginsClassLoader(ClassLoader parent){
        super(parent);
    }
    
    @Override
    public URL getResource(String name) {
        URL c = null;
        List<ClassLoader> copy = new ArrayList<ClassLoader>(this.loader);
        for (ClassLoader loader : copy) {        
            c = loader.getResource(name);
            if(c != null){
                return c;
            }
        }
        return super.getResource(name);
    }

    @Override
    public InputStream getResourceAsStream(String name) {
        InputStream c = null;
        List<ClassLoader> copy = new ArrayList<ClassLoader>(this.loader);
        for (ClassLoader loader : copy) {        
            c = loader.getResourceAsStream(name);
            if(c != null){
                return c;
            }
        }        
        return super.getResourceAsStream(name);
    }

    @Override
    public Enumeration<URL> getResources(String name) throws IOException {
        Enumeration<URL> c = null;
        List<ClassLoader> copy = new ArrayList<ClassLoader>(this.loader);        
        for (ClassLoader loader : copy) {
            try{
                c = loader.getResources(name);
                if(c.hasMoreElements()){
                    return c;
                }
            }catch(IOException e){} // ignore and continue search
        }
        return super.getResources(name);
    }

    @Override
    public Class<?> loadClass(String name) throws ClassNotFoundException {
        Class<?> c = null;
        List<ClassLoader> copy = new ArrayList<ClassLoader>(this.loader);
        for (ClassLoader loader : copy) {        
            try{
                c = loader.loadClass(name);
                if(c != null){
                    return c;
                }
            }catch(ClassNotFoundException e){} // ignore and continue searching                            
        }        
        return super.loadClass(name);
    }

    /**
     * Register a {@link ClassLoader} that is used during class search
     * 
     * @param loader the registered loader
     */
    public void registerPluginLoader(PluginClassLoader loader) {
        this.loader.add(loader);
    }
    
    /**
     * Checks for any connected {@link URLClassLoader} isntances and extracts
     * the URLS
     * 
     * @return urls
     */
    public URL[] getCurrentURLs(){
        ArrayList<URL> urls = new ArrayList<URL>();
        List<ClassLoader> copy = new ArrayList<ClassLoader>(this.loader);
        for (ClassLoader loader : copy) {        
            if(loader instanceof URLClassLoader){
                URL[] loaderURLs = ((URLClassLoader) loader).getURLs();
                for (URL url : loaderURLs) {
                    if(!urls.contains(url))
                        urls.add(url);
                }
            }
        }
        URL[] u = new URL[urls.size()];
        u = urls.toArray(u);
        return u;
    }

    /**
     * Returns the current instance of the AllPluginsClassLoader. If the
     * loader is not initialized, this creates a new AllPluginsClassLoader using
     * {@link ClassLoader#getSystemClassLoader()} as parent.
     *
     * @return allpluginsclassloader instance
     */
    public static AllPluginsClassLoader getInstance(){
        if(instance == null){
            /*
            When this is started from within a webstart app, we have to
            use the webstart classloader, which is also the current threads
            classloader. 
             */
            ClassLoader parent = Thread.currentThread().getContextClassLoader();
            if(parent == null){
                parent = ClassLoader.getSystemClassLoader(); 
            }
            instance = new AllPluginsClassLoader(parent);
        }
        return instance;
    }
}
