/*
 * Pluginsystem
 * Copyright (C) Thasso Griebel
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package org.java.plugin.extension;

import com.google.inject.*;
import com.google.inject.matcher.Matcher;
import com.google.inject.name.Names;
import com.google.inject.util.Providers;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.java.plugin.JpfException;
import org.java.plugin.Plugin;
import org.java.plugin.PluginController;
import org.java.plugin.PluginLifecycleException;
import org.java.plugin.extension.annotations.Nullable;
import org.java.plugin.extension.annotations.Nullables;
import org.java.plugin.extension.converter.Converter;
import org.java.plugin.extension.converter.ConverterGraph;
import org.java.plugin.extension.converter.ExtendedConverter;
import org.java.plugin.extension.converter.TypeConverter;
import org.java.plugin.extension.util.PluginTypeConverter;
import org.java.plugin.extension.util.TypeMatchers;
import org.java.plugin.registry.Extension;
import org.java.plugin.registry.ExtensionPoint;
import org.java.plugin.registry.ParameterMultiplicity;
import org.java.plugin.registry.PluginDescriptor;

import java.lang.reflect.Type;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * This class creates instances from Extensions. It is based on Guice
 * to allow dependecy injection of Objects and parameters.
 *  
 *
 * @author Thasso Griebel (Thasso.Griebel@gmail.com)
 */
@Singleton
public class InstanceFactory {
    /**
     * null provider
     */
    private static final Provider NULL_PROVIDER = new Provider(){
        public Object get() {
            return null;
        }
    };

    /**
     * List of classes excluded from beeing converted using the plugin converter. guice
     * allready provides converters for these classes and we can not override
     */
    private final static Map<Class, Boolean> EXCLUDED_CONVERTERS = new HashMap<Class, Boolean>();
    static{
        EXCLUDED_CONVERTERS.put(Boolean.class, false);
        EXCLUDED_CONVERTERS.put(Character.class, false);
        EXCLUDED_CONVERTERS.put(Byte.class, false);
        EXCLUDED_CONVERTERS.put(Short.class, false);
        EXCLUDED_CONVERTERS.put(Integer.class, false);
        EXCLUDED_CONVERTERS.put(Long.class, false);
        EXCLUDED_CONVERTERS.put(Float.class, false);
        EXCLUDED_CONVERTERS.put(Double.class, false);
        EXCLUDED_CONVERTERS.put(Class.class, false);
        EXCLUDED_CONVERTERS.put(Enum.class, false);
    }
    @Inject
    private static InstanceFactory instance;
    /**
     * logger
     */
    private final Log log = LogFactory.getLog(InstanceFactory.class);
    /**
     * The current main injector
     */
    private Injector injector;
    /**
     * The plugin mananger
     */
    private final PluginController controller;
    /**
     * The converter graph
     */
    private ConverterGraph converterGraph;
    /**
     * Store extension injectors here
     */
    private final Map<Extension, Injector> childInjectors;
    /**
     * Store classes that have a default injector assignd
     */
    private final Map<Class, Boolean> boundClasses;

    /**
     * If true, accessing an extension class or createing an extension instance
     * will automatically trigger plugin activation
     */
    private boolean autoActivate = true;


    /**
     * Creates a new InstanceFactory where the Injector is used as
     * the main injector to create extensions (instances from extensions
     * that provide a "class" parameter) defined in the given {@link PluginController}
     *
     * @param injector the main Injector
     * @param controller the PluginManager
     * @see PluginController
     * @see com.google.inject.Injector
     * @see org.java.plugin.registry.Extension#createInstance() 
     */
    @Inject
    public InstanceFactory(Injector injector, PluginController controller, ConverterGraph converterGraph){
        super();
        this.injector = injector;
        this.controller = controller;
        this.converterGraph = converterGraph;
        childInjectors= new HashMap<Extension, Injector>();
        boundClasses = new HashMap<Class, Boolean>();
        this.controller.addPluginLifecycleListener(new ActivatePluginListener());
    }

    /**
     * Get converter extension and add them to the converter graph
     */
    public void initiazlieFromController(){
        /*
        configure converter graph
         */
        // initialize the default converters from internal plugin registration
        // the point is PluginSystem@Converter
        try {
            Plugin internal = controller.getPlugin("PluginSystem");
            ExtensionPoint converterPoint = internal.getExtensionPoint("Converter");
            for (Extension extension : converterPoint.getAvailableExtensions()) {
                try {
                    Converter c = (Converter) extension.createInstance();
                    if(c instanceof ExtendedConverter){
                        ExtendedConverter ex = (ExtendedConverter) c;
                        converterGraph.registerConverter(ex.getSource(), ex.getTarget(), c);
                    }else{
                        Type source = TypeConverter.getType(extension.get("source"));
                        Type target = TypeConverter.getType(extension.get("target"));
                        converterGraph.registerConverter(source, target, c);
                    }
                } catch (JpfException e) {
                    log.error("Unable to create Converter from " + extension, e);
                }
            }
        } catch (PluginLifecycleException e) {
            log.error("Unable to initialize converter graph from extensions");
        }

    }

    /**
     * Registers a new Type converter. This has to be called befor any opbject creation, otherwise
     * the type converter will not be added to the extension modules. Add you custom
     * converters directly after initializing the PluginManager.
     * <P>
     * You can also use this to overwrite already registered converter but overriding the Guice
     * default (such as String) does not work.
     *
     * @param targetType the target type
     * @param converter the converter class
     */
    public void registerTypeConverter(Type targetType, Converter<String, ?> converter){
        converterGraph.registerConverter(String.class, targetType, converter);
    }

    /**
     * Get or create an injector for a given Extension
     *
     * @param extension the extension
     * @return injector that can be used to create instances for the given extension
     */
    Injector getInjector(Extension extension){
        Injector child = childInjectors.get(extension);
        if(child == null){
            ExtensionModule em = new ExtensionModule(extension);
            child = injector.createChildInjector(em);
            childInjectors.put(extension, child);
        }
        return child;
    }

    /**
     * Creates an instance of the given class using the current injector.
     *
     * @param clazz to create
     * @param <T> the class' type
     * @return instance of given class
     */
    public <T> T get(Class<? extends T> clazz){
        return injector.getInstance(clazz);
    }

    /**
     * Use this to craete an instance from the given extension. The extension has to
     * provide a parameter called "class" and the class has to be loadable by the
     * declaring plugins ClassLoader. If {@link #autoActivate} is enabled, this will make sure that
     * the decalaring plugin of the given Extension is activated before access to
     * its plugin classloader.
     *
     * @param extension the
     * @return  object instance created from extension's class paramter
     * @throws ClassNotFoundException if the extensions class could not be found
     */
    public Object createFromExtension(Extension extension) throws JpfException {
        try{
            if(extension.get("class") != null){
                checkAndActivate(extension);
                Injector child = getInjector(extension);
                Class<?> c = loadClass(extension);
                return child.getInstance(c);
            }
        }catch(Exception e){
            throw new InstanceFactoryException("Unable to create instance for extension " + extension, e);
        }
        throw new InstanceFactoryException("Extension " + extension.getUniqueId() +
                " does not provide a paramter named 'class' ! ", null);
    }

    /**
     * Looks for a parameter named class and uses the plugins classloader
     * to load that class. If {@link #autoActivate} is enabled, this will make sure that
     * the decalaring plugin of the given Extension is activated before access to
     * its plugin classloader.
     *
     * @param extension the extension
     * @return class loaded from parameter class
     * @throws ClassNotFoundException if the specified exception class could not be found
     */
    public Class<?> loadClass(Extension extension) throws ClassNotFoundException {
        String className = extension.get("class");
        if(className != null){
            checkAndActivate(extension);
            ClassLoader cl = getPluginClassLoader(extension);
            return cl.loadClass(className);
        }
        throw new RuntimeException("Extension " + extension.getUniqueId() +
                " does not provide a paramter named 'class' ! ");
    }

    /**
     * Method to resolved plugin's classloader from a given extension
     *
     * @param extension the extension
     * @return plugin classloader
     */
    public ClassLoader getPluginClassLoader(Extension extension) {
        return controller.getJPFManager().getPluginClassLoader(extension.getDeclaringPluginDescriptor());
    }

    /**
     * Static access to the instance factory. WARNING Do not use this unless
     * you initialized framework properly. This class depends on Guice DI and its
     * Injector!
     *
     *
     * @return InstanceFactory
     */
    public static InstanceFactory getInstance(){
        if(instance == null){
            throw new NullPointerException("InstanceFactory was not initialized properly !");
        }
        return instance;
    }

    /**
     * Internal use: check if autoActivate is enabled. If so, check if the decalring plugin of
     * the given extension is disabled and activate it eventually.
     *
     * @param extension the extension
     */
    private void checkAndActivate(Extension extension) {
        if(autoActivate){
            PluginDescriptor desc = extension.getDeclaringPluginDescriptor();
            if(! controller.getJPFManager().isPluginActivated(desc)
                    && ! controller.getJPFManager().isPluginActivating(desc)){
                try {
                    controller.activatePlugin(desc);
                } catch (PluginLifecycleException e) {
                    log.error("Error while activating plugin...trying to continue!", e);
                }
            }
        }
    }

    /**
     *
     * @return true if auto activation is enabled
     */
    public boolean isAutoActivate() {
        return autoActivate;
    }

    /**
     * Use this to enable/disable automatic plugin activation.
     * If this is true, plugins will be automatically activated before the InstanceFactory
     * acesses an extension or an extension class.
     *  
     * @param autoActivate enabled
     * @see #loadClass(org.java.plugin.registry.Extension)
     * @see #createFromExtension(org.java.plugin.registry.Extension) 
     */
    public void setAutoActivate(boolean autoActivate) {
        this.autoActivate = autoActivate;
    }

    /**
     * Helper method to access the converter facility and convert from source to target type
     *
     * @param value the value
     * @param source the source type
     * @param target the target type
     * @param loader the loader
     * @param <T> target type
     * @return converted value or null
     */
    public <T> T convert(Object value, Type source, Type target, ClassLoader loader) {
        Converter c = converterGraph.getConverter(source, target);
        if ( c == null ) return null;
        return (T) c.convert(value, loader != null ? loader: AllPluginsClassLoader.getInstance());
    }

    /**
     * Helper to initialize extension on plugin activation
     */
    private class ActivatePluginListener implements PluginLifecycleListener{
        public void pluginActivated(Plugin plugin) {
            ArrayList<Module> modules = new ArrayList<Module>();
            List<Module> pluginModules = plugin.getModules();
            if(pluginModules != null){
                modules.addAll(pluginModules);
            }
            for (Extension extension : plugin.getExtensions()) {
                /*
                make sure we create the child injector BEFORE we create the new
                main injector. This ensures that the child injector does NOT have a binding
                to itself (binding the extension class to its provider)
                 */
                getInjector(extension);
                if(extension.getParameter("class") != null){
                    modules.add(new ExtensionProviderModule(extension));
                }
            }
            if(modules.size() > 0){
                injector = injector.createChildInjector(modules);
            }
        }
        public void pluginDeactivated(Plugin plugin) {}
        public void pluginDisabled(PluginDescriptor descriptor) {}
        public void pluginEnabled(PluginDescriptor descriptor) {}
    }

    /**
     * A {@link com.google.inject.Module} for a given Extension. The modules context contains
     * constant bindings to the extension parameter and all typeConverters are added.
     */
    class ExtensionModule implements Module {
        private TypeLiteral<List<String>> stringListType = new TypeLiteral<List<String>>() {};
        private final Extension extension;


        /**
         * Create a new Extension module for a given {@link org.java.plugin.registry.Extension}
         *
         * @param extension the extension
         */
        public ExtensionModule(Extension extension){
            this.extension = extension;
        }
        public void configure(Binder binder) {
            HashMap<String, String> props = new HashMap<String, String>();
            ExtensionPoint ep = controller.getJPFManager()
                                .getRegistry()
                                .getExtensionPoint(extension.getExtendedPluginId(),
                                                   extension.getExtendedPointId());
            // To implement #3 we have to check the parameter multiplicity and create
            // lists of parameters for any none-or-more and one-or-more parameters
            // also we use the extension point to get parameters and set default values
            // if the extension does not provide the parameter
            for (ExtensionPoint.ParameterDefinition definition : ep.getParameterDefinitions()) {
                if(definition.getMultiplicity() == ParameterMultiplicity.ONE || definition.getMultiplicity() == ParameterMultiplicity.NONE_OR_ONE){
                    // one parameter has to be set in the extension point
                    // but we check for a default anyways and throw an exception
                    // only of no parameter and no default is set
                    if(extension.getParameter(definition.getId()) != null){
                        props.put(definition.getId(), extension.getParameter(definition.getId()).rawValue());
                    }else{
                        // check for default
                        if(definition.getDefaultValue() != null){
                            props.put(definition.getId(), definition.getDefaultValue());
                        }else{
                            if(definition.getMultiplicity()== ParameterMultiplicity.ONE){
                                throw new RuntimeException("Parameter " + definition.getId() + " is not set in extension " + extension + " and the ExtensionPoint does not provide a default value!");
                            }else{
                                // bind to null for NONE_OR_ONE multiplicity
                                props.put(definition.getId(), null);
                            }
                        }
                    }
                }else{
                    // bind to collection
                    List<String> values = new ArrayList<String>();
                    for (Extension.Parameter parameter : extension.getParameters(definition.getId())) {
                        values.add(parameter.rawValue());
                    }
                    // we bind also the empty list only if the multiplicity is ANY
                    if(values.size() > 0 || definition.getMultiplicity() == ParameterMultiplicity.ANY){
                        // we do not use the hashmap here and rather bind directly as
                        // can only bind to list of strings
                        binder.bind(stringListType).annotatedWith(Names.named(definition.getId())).toInstance(values);
                    }else{
                        // if there should be a defined parameter,we throw an exception here
                        if(definition.getMultiplicity() == ParameterMultiplicity.ONE_OR_MORE){
                            throw new RuntimeException("Parameter " + definition.getId() + " is not set in extension " + extension + " and the ExtensionPoint does not provide a default value!");
                        }
                    }

                }
            }

            
            for (Extension.Parameter p : extension.getParameters()) {
                if(p.rawValue() != null){                    
                    props.put(p.getId(), p.rawValue());
                }
            }
//            Names.bindProperties(binder, props);
//            for (Object o : props.keySet()) {
//                Object value =
//                binder.bind(Key.get(String.class, new NamedImpl(key))).annotatedWith(Names.named(o.toString())).to((String) props.get(o));
//                binder.bindConstant().annotatedWith(Nullable.class).to
//            }

            for (Map.Entry<String, String> entry : props.entrySet()) {
              String key = entry.getKey();
              String value = entry.getValue();
             if(value != null){
                binder.bind(Key.get(String.class, Names.named(key))).toInstance(value);
                binder.bind(String.class).annotatedWith(Nullables.nullable(key)).toInstance(value);
              }else{
                // bind to empty string. That way, everything that is not annotated
                // nullable, will start the converter chain and we can intercept the null
                // pointer in the PluginTypeConverter to produce better error messages
                binder.bind(Key.get(String.class, Names.named(key))).toInstance("");
                binder.bind(String.class).annotatedWith(Nullables.nullable(key)).toProvider(Providers.of((String)null));
              }
            }

            // gat target types for String and add all interfaces and super types
            List<Type> targetTypes = converterGraph.getTargetTypes(String.class);
            List<Class> classes = new ArrayList<Class>();
            for (Type targetType : targetTypes) {
                if(targetType instanceof Class){
                    Class c = (Class) targetType;
                    if(!classes.contains(c))
                        classes.add(c);
                    for (Class interfaceClass : c.getInterfaces()) {
                        if(!classes.contains(interfaceClass)){
                            classes.add(interfaceClass);
                        }
                    }


                    Class superClass = c.getSuperclass();
                    while(superClass != null){
                        if(!classes.contains(superClass))
                            classes.add(superClass);
                        for (Class interfaceClass : superClass.getInterfaces()) {
                            if(!classes.contains(interfaceClass)){
                                classes.add(interfaceClass);
                            }
                        }
                        superClass = superClass.getSuperclass();
                    }
                }
            }

            for (Class type : classes) {
                Class<?> c = (Class<?>) type;
                binder.bind(c).annotatedWith(Nullable.class).toProvider(NULL_PROVIDER);
            }


            ClassLoader pluginLoader = getPluginClassLoader(extension);
            PluginTypeConverter converter = new PluginTypeConverter(pluginLoader, converterGraph);



            // remove all superclasses because we use the superclass matcher
            // and remove all excluded that already have a guice converter
            List<Type> types = converterGraph.getTargetTypes();
            ArrayList<Type> toRemove = new ArrayList<Type>();
            for (Type type : types) {
                Matcher m = TypeMatchers.superclassesOf(TypeLiteral.get(type));
                for (Type t : types) {
                    if(t == type) continue;
                    if(m.matches(TypeLiteral.get(t))){
                        toRemove.add(t);
                    }else if(t instanceof Class && EXCLUDED_CONVERTERS.containsKey((Class)t)){
                        toRemove.add(t);
                    }
                }
            }

            types.removeAll(toRemove);
            for (Type type : types) {
                TypeLiteral tl = TypeLiteral.get(type);
                binder.convertToTypes(TypeMatchers.superclassesOf(tl), converter);               
            }

        }
    }

    /**
     * This module is used to buid the next injector in the injector hierarchy.
     * It binds the Class specified in the extension to a custom provider.
     */
    class ExtensionProviderModule extends AbstractModule{
        private final Extension extension;

        public ExtensionProviderModule(Extension extension){
            this.extension = extension;
        }

        protected void configure() {
            try {
                final Class<?> c = loadClass(extension);
                Provider provider = new Provider() {
                    public Object get() {
                        return getInjector(extension).getProvider(c).get();
                    }
                };
                /*
                bind annotated version
                the injection point has to be annotated with name == extension.uniqueID
                i.e:
                @Inject @Named("somePlugin@someExtensionID") ...
                 */
                Key<?> key = Key.get(c, Names.named(extension.getUniqueId()));
                bind(key).toProvider(provider);

                /*
                 * if the extension class is not bound yet, use this as
                 * the default extension
                 */
                if(!boundClasses.containsKey(c)){
                    bind(Key.get(c)).toProvider(provider);
                    boundClasses.put(c, true);
                }
            } catch (ClassNotFoundException e) {
                e.printStackTrace();
            }
        }
    }

    /**
     * Exception thrown but the InstanceFactory
     */
    class InstanceFactoryException extends JpfException{

        protected InstanceFactoryException(final String message, final Throwable cause) {
            super(message, cause);
        }
    }
}
