/*
 * Pluginsystem
 * Copyright (C) Thasso Griebel
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package org.java.plugin.extension;


import java.io.File;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URL;
import java.util.*;

import javax.xml.parsers.SAXParserFactory;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.java.plugin.PluginManager.PluginLocation;
import org.java.plugin.standard.StandardPluginLocation;
import org.xml.sax.Attributes;
import org.xml.sax.helpers.DefaultHandler;

/**
 * 
 * This boroughs code from the JPF boot library with a few modifications.
 * <br>
 * Default implementation of plug-ins collector. Supported
 * configuration parameters are:
 * <dl>
 *   <dt>Application.pluginRepositories</dt>
 *   <dd>Comma separated list of local plug-in repositories, given folders will
 *     be scanned for plug-ins. Default value is <code>./plugins</code>.</dd>
 *   <dt>epos.pluginLocationDescriptors</dt>
 *   <dd>Comma separated list of URLs for XML syntax files that describe
 *     available plug-in locations (see file syntax bellow). No default value
 *     provided.</dd>
 * </dl>
 * <p>
 * Given repositories are scanned recursively collecting all folders that
 * contain <code>plugin.xml</code> or <code>plugin-fragment.xml</code> and
 * <code>*.zip</code> and <code>*.jar</code> files.
 * </p>
 * <p>
 * Plug-ins locations descriptor is a simple XML syntax file that stores
 * locations of all available plug-in manifests and contexts (in terms of
 * {@link org.java.plugin.PluginManager.PluginLocation}). Here is an example:
 * <pre>&lt;plugins&gt;
 *   &lt;plugin
 *     manifest="http://localhost/myPlugins/plugin1/plugin.xml"
 *     context="http://localhost/myPlugins/plugin1/"/&gt;
 *   &lt;plugin
 *     manifest="http://localhost/myPlugins/plugin2/plugin.xml"
 *     context="http://localhost/myPlugins/plugin2/"/&gt;
 *   &lt;plugin
 *     manifest="http://www.plugins.com/repository/plugin1/plugin.xml"
 *     context="http://www.plugins.com/repository/plugin1/"/&gt;
 *   &lt;plugin
 *     manifest="http://www.plugins.com/repository/plugin1/plugin.xml"
 *     context="http://www.plugins.com/repository/plugin1/"/&gt;
 * &lt;/plugins&gt;</pre>
 * Using such simple descriptor you may, for example, publish plug-ins on a site
 * to make them available for clients without needing to download plug-ins
 * manually.
 * </p>
 */
public class PluginsCollector {
    
    protected static final String PARAM_PLUGINS_REPOSITORIES = "jpf.plugindir";
    protected static final String PARAM_PLUGINS_LOCATIONS_DESCRIPTORS ="jpf.plugin.descriptors"; 

    protected Log log = LogFactory.getLog(this.getClass());
    
    /**
     * Local Plugin locations
     */
    private List<File> repositories;
    /**
     * Descriptors
     */
    private List<URL> descriptors;
    /**
     * Plugin jar files
     */
    private List<URL> jars;
         
    /**
     * Create and configure a new PluginsCollector.
     * 
     * @param configuration
     */
    public PluginsCollector(Properties configuration){
    	super();
    	configure(configuration);
    }
    /**
     * Takes the configuration from the given properties and scans for plugins and description
     * locations. This does not create {@link PluginLocation}s but detects local directories and
     * remote {@link URL}s.
     * 
     * @param config
     * @throws Exception
     */
    protected void configure(Properties config) {    	
        repositories = new LinkedList<File>();
        for (StringTokenizer st = new StringTokenizer(
                config.getProperty(PARAM_PLUGINS_REPOSITORIES,
                        '.' + File.separator + "plugins"),
                        ",", false); st.hasMoreTokens();) {        	
            String token = st.nextToken().trim();            
            if (token.length() == 0) {
                continue;
            }
            try{
                File file = null;
                file = new File(token).getCanonicalFile();
                if(file != null)repositories.add(file);
            }catch(Exception e){
                log.warn("plugindirecory " + token + " not found - skipping");
            }
        }
        // check for plugin.xml or META-INF/plugin.xml file in the classpath
        List<URL> localClassloaderURLs = new ArrayList<URL>();
        List<URL> systemClassloaderURLs  = new ArrayList<URL>();

        try{
            Enumeration<URL> urlEnumeration = getClass().getClassLoader().getResources("plugin.xml");
            if(urlEnumeration != null){
                while (urlEnumeration.hasMoreElements()){
                    localClassloaderURLs.add(urlEnumeration.nextElement());
                }
            }
            urlEnumeration = getClass().getClassLoader().getResources("META-INF/plugin.xml");
            if(urlEnumeration != null){
                while (urlEnumeration.hasMoreElements()){
                    localClassloaderURLs.add(urlEnumeration.nextElement());
                }
            }
        }catch(Exception e){}
        try{
            Enumeration<URL> urlEnumeration = ClassLoader.getSystemResources("plugin.xml");
            if(urlEnumeration != null){
                while (urlEnumeration.hasMoreElements()){
                    systemClassloaderURLs.add(urlEnumeration.nextElement());
                }
            }
            
            urlEnumeration = ClassLoader.getSystemResources("META-INF/plugin.xml");
            if(urlEnumeration != null){
                while (urlEnumeration.hasMoreElements()){
                    systemClassloaderURLs.add(urlEnumeration.nextElement());
                }
            }

        }catch(Exception e){}

        Set<URL> jarDirs = new HashSet<URL>();
        jars = new ArrayList<URL>();

        if(localClassloaderURLs != null){
        	log.debug("Found localClassloader urls " +(localClassloaderURLs.size() > 0));
            for (URL url : localClassloaderURLs) {
                jarDirs.add(jarelement2jar(url));
            }
        }else{
        	log.debug("Found localClassloder urls " + 0);
        }
        if(systemClassloaderURLs != null){
        	log.debug("Found systemClassloader urls " +(systemClassloaderURLs.size()>0));
            for (URL url : systemClassloaderURLs) {
                jarDirs.add(jarelement2jar(url));
            }
        }else{
        	log.debug("Found systemClassloder urls " +0);
        }
        
        for (URL u : jarDirs) {
        	log.debug("Adding url from classpath jar " + u);
			jars.add(u);
		}
                
        log.debug("found " + repositories.size() + " local plugin repositories");
        
        descriptors = new LinkedList<URL>();
        for (StringTokenizer st = new StringTokenizer(
                config.getProperty(PARAM_PLUGINS_LOCATIONS_DESCRIPTORS, ""), 
                ",", false); st.hasMoreTokens();) {
            String token = st.nextToken().trim();
            if (token.length() == 0) {
                continue;
            }
            try{descriptors.add(new URL(token));}catch(Exception e){}
        }
        log.debug("found " + descriptors.size()+ " plugin location descriptors");        
    }

    /**
     * After {@link #configure(Properties)} was called, this scans the found 
     * directories and locations and creates an array of {@link PluginLocation}s. 
     */
    public PluginLocation[] collectPluginLocations() {
        List<PluginLocation> result = new LinkedList<PluginLocation>();
        for (File file : repositories) {			            
            if (file.isDirectory()) {
                processFolder(file, result);
            } else if (file.isFile()) {
                processFile(file, result);
            } else {
                log.warn("unknown repository location " + file);
            }
        }
        for (URL url : descriptors) {
        	processDescriptor(url, result);
		}
        for (URL u : jars) {
			PluginLocation l = new StandardPluginLocation(u,u);
			result.add(l);
		}
        return result.toArray(new PluginLocation[result.size()]);
    }

    /**
     * Process a folder while creating {@link PluginLocation}s.
     * This creates a {@link PluginLocation} for the folder and recursively for each sub-folder. 
     * 
     * @param folder the folder
     * @param result list of {@link PluginLocation}
     */
    protected void processFolder(final File folder, final List<PluginLocation> result) {
        log.debug("processing folder - " + folder);
        try {
            PluginLocation pluginLocation =
                StandardPluginLocation.create(folder);
            if (pluginLocation != null) {
                result.add(pluginLocation);
                return;
            }
        } catch (MalformedURLException mue) {
            log.warn("failed collecting plug-in folder " + folder 
                    + ", ignoring it", mue);
            return;
        }
        File[] files = folder.listFiles();
        for (int i = 0; i < files.length; i++) {
            File file = files[i];
            if (file.isDirectory()) {
                processFolder(file, result);
            } else if (file.isFile()) {
                processFile(file, result);
            }
        }
    }
    /**
     * Process a file as {@link PluginLocation}. This creates a {@link PluginLocation} for the given file.
     * 
     * @param file
     * @param result
     */
    protected void processFile(final File file, final List<PluginLocation> result) {
        log.debug("processing file - " + file);
        try {
            PluginLocation pluginLocation = StandardPluginLocation.create(file);
            if (pluginLocation != null) {
                result.add(pluginLocation);
            }
        } catch (MalformedURLException mue) {
            log.warn("failed collecting plug-in file " + file
                    + ", ignoring it", mue);
        }
    }
    
    private void processDescriptor(final URL url, final List<PluginLocation> result) {
        log.debug("processing plug-ins locations descriptor, URL=" + url);
        try {
            SAXParserFactory.newInstance().newSAXParser().parse(
                    url.toExternalForm(),
                    new LocationsDescriptorHandler(result));
        } catch (Exception e) {
            log.warn("failed processing plug-ins locations descriptor, URL="
                    + url, e);
        }
    }

    private final class LocationsDescriptorHandler extends DefaultHandler {
        private final List<PluginLocation> resultData;

        LocationsDescriptorHandler(final List<PluginLocation> result) {
            resultData = result;
        }
        
        /**
         * @see org.xml.sax.helpers.DefaultHandler#startElement(
         *      java.lang.String, java.lang.String, java.lang.String,
         *      org.xml.sax.Attributes)
         */
        public void startElement(final String uri, final String localName,
                final String qName, final Attributes attributes) {
            if (!"plugin".equals(qName)) { //$NON-NLS-1$
                return;
            }
            String manifest = attributes.getValue("manifest"); //$NON-NLS-1$
            if (manifest == null) {
                log.warn("manifest attribute missing"); //$NON-NLS-1$
                return;
            }
            URL manifestUrl;
            try {
                manifestUrl = new URL(manifest);
            } catch (MalformedURLException mue) {
                log.warn("invalid manifest URL - " + manifest, mue); //$NON-NLS-1$
                return;
            }
            String context = attributes.getValue("context"); //$NON-NLS-1$
            if (context == null) {
                log.warn("context attribute missing"); //$NON-NLS-1$
                return;
            }
            URL contextUrl;
            try {
                contextUrl = new URL(context);
            } catch (MalformedURLException mue) {
                log.warn("invalid context URL - " + context, mue); //$NON-NLS-1$
                return;
            }
            resultData.add(new StandardPluginLocation(contextUrl, manifestUrl));
            log.debug("got plug-in location from descriptor, manifestUrl=" //$NON-NLS-1$
                    + manifestUrl + ", contextURL=" + contextUrl); //$NON-NLS-1$
        }
    }
    
    
    protected static String jarelement2jarfile(URL cUrl){
    	/*
    	 * check working directory directory
         * detect the URL to the jar file containing Boot.class
         * we use this to detect the install directory.
         * if no jar file found, we are in some sort of development environment,
         * we assume that
         * the path will be user.dir which should be equal to current working directory
    	 */
        String absName = null;
        if (cUrl != null) {
            String c = cUrl.toString();
            if (c.startsWith("jar:")) {
            	//System.out.println("Found jar !");
                try {
                    // cut jar:
                    c = "file:///" + c.substring(9);
                    // cut all after STree.jar!
                    int index = c.lastIndexOf('!');
                    c = c.substring(0, index);
                    c = c.substring(0, c.lastIndexOf('/'));
                } catch (Exception e) {
                    e.printStackTrace();
                }
                absName = c;
                //System.out.println("Done : " + absName);
                File f = new File(URI.create(absName));
                absName = f.getAbsolutePath();
            }
        }
        return absName;
    }
    
    protected static URL jarelement2jar(URL cUrl){
    	return cUrl;
//    	/*
//    	 * check working directory directory
//         * detect the URL to the jar file containing Boot.class
//         * we use this to detect the install directory.
//         * if no jar file found, we are in some sort of development environment,
//         * we assume that
//         * the path will be user.dir which should be equal to current working directory
//    	 */
//        String absName = null;
//        if (cUrl != null) {
//            String c = cUrl.toString();
//            if (c.startsWith("jar:")) {
//            	System.out.println("Found jar !" + c);
//                try {
//                    // cut jar:
//                    c = "file:///" + c.substring(9);
//                    // cut all after STree.jar!
//                    int index = c.lastIndexOf('!');
//                    c = c.substring(0, index);
//                    //c = c.substring(0, c.lastIndexOf('/'));
//                } catch (Exception e) {
//                    e.printStackTrace();
//                }
//                absName = c;
//                System.out.println("Done : " + absName);
//                File f = new File(URI.create(absName));
//                absName = f.getAbsolutePath();
//            }
//        }
//        return absName;
    }

    /**
     *
     * @return repositories registered
     */
    public List<File>  getRepositories(){
        return Collections.unmodifiableList(repositories);
    }

    /**
     *
     * @return descriptors registered
     */
    public List<URL>  getDescriptors(){
        return Collections.unmodifiableList(descriptors);
    }

    /**
     *
     * @return jars registered
     */
    public List<URL>  getJars(){
        return Collections.unmodifiableList(jars);
    }

    /**
     * Add a new location as plugin repository. This checks if the location
     * already exists in this collector. If not, it is added. If
     * you add repositories here, make sure to register them with
     * your plugin manager.
     *
     *
     * @param location the new location
     * @return true if location was added
     */
    public boolean addRepository(File location) {
        if(location == null) return false;
        if(repositories == null){
            throw new RuntimeException("PluginsCollector is not initialized. Initialize the collector before you add repositories manually.");
        }
        // check contains. Because of the absolute vs relative paths
        // we can not use file.equals,which uses abstract path names
        for (File repository : repositories) {
            if(repository.getAbsoluteFile().equals(location.getAbsoluteFile())){
                return false;
            }
        }
        if(!repositories.contains(location)){
            return repositories.add(location);
        }
        return false;        
    }



}
