/*
 * Pluginsystem
 * Copyright (C) Thasso Griebel
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package org.java.plugin.extension.converter;

import org.jgrapht.DirectedGraph;
import org.jgrapht.EdgeFactory;
import org.jgrapht.event.TraversalListener;
import org.jgrapht.event.TraversalListenerAdapter;
import org.jgrapht.event.EdgeTraversalEvent;
import org.jgrapht.traverse.DepthFirstIterator;
import org.jgrapht.alg.DijkstraShortestPath;
import org.jgrapht.graph.DefaultDirectedGraph;
import org.java.plugin.extension.util.TypeMatchers;

import java.lang.reflect.Type;
import java.util.List;
import java.util.ArrayList;
import java.util.Collections;

import com.google.inject.Singleton;
import com.google.inject.TypeLiteral;
import com.google.inject.matcher.Matcher;

/**
 * This class can be used to find instances of {@link org.java.plugin.extension.converter.Converter} that
 * can be used to convert one data type into the other.
 * <p>
 * The current implementation is stable for classes but will not work right for Collections.
 * <p>
 * The implementations builds a directed graph where edges are Converter and Vertices are the Types. To
 * find a proper converter, we do a shortest path search (using Dijkstras algorithm). If the path length
 * is {@literal > 1 } we combine the converter into a converter pipeline.
 * 
 * <p>
 * WARNING : This is under development and the API might change
 * @author Thasso Griebel (thasso.griebel@gmail.com)
 */
@Singleton
public class ConverterGraph {
    /**
     * The underlying graph
     */
    private DirectedGraph<Type, Converter> graph;

    public ConverterGraph(){
        super();
        graph = new DefaultDirectedGraph<Type, Converter>(new ConverterEdgeFactory());
    }
    
    public boolean registerConverter(Type sourceType, Type targetType, Converter converter){
        // check for primitive
        sourceType = convertToWrapper(sourceType);
        targetType = convertToWrapper(targetType);
        
        if(graph.containsEdge(sourceType, targetType)) return false;
        if(!graph.containsVertex(sourceType)) graph.addVertex(sourceType);
        if(!graph.containsVertex(targetType)) graph.addVertex(targetType);
        graph.addEdge(sourceType, targetType, converter);

        // add edges between target type and all existing super types of the target type
        // assume you have
        // URL->BufferedImage
        // Image->ImageIcon
        //
        // this implies :
        // BufferedImage->Image
        if(targetType instanceof Class){
            Class targetClass = (Class) targetType;
            for (Type type : graph.vertexSet()) {
                if(type instanceof Class){ // raw type
                    Class typeClass = (Class) type;
                    if(typeClass.isAssignableFrom(targetClass)){
                        graph.addEdge(targetType, type,new SupertypeConverter());
                    }else if(targetClass.isAssignableFrom(typeClass)){
                        graph.addEdge(type, targetType, new SupertypeConverter());
                    }
                }
            }
        }
        // same for sources
        if(sourceType instanceof Class){
            Class sourceClass = (Class) sourceType;
            for (Type type : graph.vertexSet()) {
                if(type instanceof Class){ // raw type
                    Class typeClass = (Class) type;
                    if(typeClass.isAssignableFrom(sourceClass)){
                        graph.addEdge(sourceType, type,new SupertypeConverter());
                    }else if(sourceClass.isAssignableFrom(typeClass)){
                        graph.addEdge(type, sourceType, new SupertypeConverter());
                    }
                }
            }
        }

        return true;
    }

    /**
     * Converts the given type to the appropriate Wrapper class
     * if type is a Class and is primitive.
     *
     * @param t the type
     * @return type that is either the wrapper class or the original type if that was no primitive
     */
    private Type convertToWrapper(Type t){
        if(t instanceof Class){
            Class c = (Class) t;
            if(c.isPrimitive()){
                if(c == Boolean.TYPE)return Boolean.class;
                else if(c== Character.TYPE) return Character.class;
                else if(c== Byte.TYPE) return Byte.class;
                else if(c== Short.TYPE) return Short.class;
                else if(c== Integer.TYPE) return Integer.class;
                else if(c== Long.TYPE) return Long.class;
                else if(c== Float.TYPE) return Float.class;
                else if(c== Double.TYPE) return Double.class;
                else if(c== Void.TYPE) return Void.class;
            }
        }
        return t;
    }


    protected boolean isTarget(Type t){
        return graph.containsVertex(t) && graph.inDegreeOf(t)>0;   
    }

    protected boolean isSource(Type t){
        return graph.containsVertex(t) && graph.outDegreeOf(t)>0;   
    }

    protected Type toClassTarget(Class t){
        for (Type type : graph.vertexSet()) {
            if(type instanceof Class){
                Class typeClass = (Class) type;
                if(t.isAssignableFrom(typeClass)){
                    return typeClass;
                }
            }
        }
        return null;
    }

    protected Type toClassSource(Class t){
        Class typeClass = (Class) t;
        while(typeClass.getSuperclass() != null){
            typeClass = typeClass.getSuperclass();
            if(isSource(typeClass)) return typeClass;
        }
        return null;
    }


    public Converter getConverter(Type source, Type target){
        Type s = source;
        Type t = target;

        // check for primitive
        s = convertToWrapper(s);
        t = convertToWrapper(t);


        if(!isSource(s)){
            if(s instanceof Class){
                s = toClassSource((Class) s);
                if(s == null) return null;
            }
        }
        if(!isTarget(t)){
            if(t instanceof Class){
                t = toClassTarget((Class) t);
                if(t == null) return null;
            }
        }

        Converter converter = graph.getEdge(s, t);
        if(converter == null){
            List<Converter> converters = DijkstraShortestPath.findPathBetween(graph, s, t);
            if(converters != null){
                if(converters.size() == 1){
                    converter = converters.get(0);
                }else{
                    converter = new PipedConverter(converters);
                }
            }
        }
        return converter;
    }

    /**
     * Returns all target types. That is, all types with indegree {@literal > 0}.
     *
     * @return nodes with indegree {@literal > 1}
     */
    public List<Type> getTargetTypes() {
        ArrayList<Type> types = new ArrayList<Type>();
        for (Type t : graph.vertexSet()) {
            if(graph.inDegreeOf(t) > 0){
                if(!types.contains(t)){
                    types.add(t);
                }
            }
        }
        return types;
    }

    public List<Type> getTargetTypes(Type source) {
        if(!graph.containsVertex(source)) return Collections.emptyList();
        final ArrayList<Type> types = new ArrayList<Type>();
        DepthFirstIterator<Type, Converter> dfi = new DepthFirstIterator<Type, Converter>(graph, source);
        while(dfi.hasNext()){
            types.add(dfi.next());
        }        
        return types;
    }



    class ConverterEdgeFactory implements EdgeFactory{
        public Object createEdge(Object o, Object o1) {
            return null;
        }
    }

    class SupertypeConverter implements Converter{
        public Object convert(Object o, ClassLoader loader) {
            return o;
        }
    }
}
