/*
 * Pluginsystem
 * Copyright (C) Thasso Griebel
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package org.java.plugin.extension.converter;

import org.java.plugin.extension.annotations.Extension;

import javax.imageio.ImageIO;
import javax.swing.*;
import java.net.URL;
import java.net.MalformedURLException;
import java.io.IOException;
import java.io.File;
import java.awt.*;
import java.awt.image.BufferedImage;
import java.lang.reflect.Type;
import java.util.logging.Logger;
/*
* Date: Jun 10, 2009
*/

/**
 * Collection of default converters
 * 
 * @author Thasso Griebel (thasso.griebel@gmail.com)
 */
public class Converters {
    /**
     * Converts a String to an URL. If the string contains the protocol (identified by
     * searching for "://"), this just creates the URL. Otherwise this uses the given loader
     * to resolve the string as resoruce in the classpath.
     */
    @Extension("PluginSystem@Converter")
    public static class String2Url extends StringConverter<URL>{
        public URL convert(String s, ClassLoader loader) {
            if(s.contains("://")){
                try {
                    return new URL(s);
                } catch (MalformedURLException e) {
                    throw new RuntimeException(e);
                }
            }
            return loader.getResource(s);
        }

        public Type getTarget() {
            return URL.class;
        }
    }

    /**
     * Convert from a string to a file using the files string constructor
     */
    @Extension("PluginSystem@Converter")
    public static class String2File extends StringConverter<File>{
        public File convert(String s, ClassLoader loader) {
            return new File(s);
        }

        public Type getTarget() {
            return File.class;
        }
    }
    
    /**
     *Simple toString() based object to string converter
     */
    @Extension("PluginSystem@Converter")
    public static class Object2String implements ExtendedConverter<Object, String> {
        public Type getSource() {
            return Object.class;
        }

        public Type getTarget() {
            return String.class;
        }

        public String convert(Object o, ClassLoader loader) {
            return o.toString();
        }
    }


    /**
     * Loads an image from the given URL  or returns null
     */
    @Extension("PluginSystem@Converter")
    public static class Url2Image implements ExtendedConverter<URL, BufferedImage> {
        public BufferedImage convert(URL imageUrl, ClassLoader loader) {
            if(imageUrl != null){
                try {
                    return ImageIO.read(imageUrl);
                } catch (IOException e) {
                    Logger.getAnonymousLogger().fine("unable to read image " + imageUrl);

                }
            }
            return null;
        }

        public Type getSource() {
            return URL.class;
        }

        public Type getTarget() {
            return BufferedImage.class;
        }
    }

    /**
     * Creates an ImageIcon from the given image
     */
    @Extension("PluginSystem@Converter")
    public static class Image2Icon implements ExtendedConverter<Image, Icon>{
        public ImageIcon convert(Image image, ClassLoader loader) {
            if(image != null){
                return new ImageIcon(image);
            }
            return null;
        }

        public Type getSource() {
            return Image.class;
        }

        public Type getTarget() {
            return ImageIcon.class;
        }
    }

    /**
     * Loads Class from stringusing given laoder
     */
    @Extension("PluginSystem@Converter")
    public static class String2Class implements ExtendedConverter<String, Class>{
        public Class convert(String s, ClassLoader loader) {
            if(loader == null) throw new NullPointerException();
            try {
                return loader.loadClass(s);
            } catch (ClassNotFoundException e) {
                Logger.getAnonymousLogger().fine("unable to load class " + s + " from  " + loader);
            }
            return null;
        }

        public Type getSource() {
            return String.class;
        }

        public Type getTarget() {
            return Class.class;
        }
    }

    /**
     * Converts String to boolean. Returns true iff {@code s.equalsIgnoreCase("true")}.
     */
    @Extension("PluginSystem@Converter")
    public static class String2Boolean implements ExtendedConverter<String, Boolean>{
        public Boolean convert(String s, ClassLoader loader) {
            if(s == null) return null;
            return s.equalsIgnoreCase("true");
        }

        public Type getSource() {
            return String.class;
        }

        public Type getTarget() {
            return Boolean.class;
        }
    }

    @Extension("PluginSystem@Converter")
    public static class String2Character implements ExtendedConverter<String, Character>{
        public Character convert(String s, ClassLoader loader) {
            if(s == null) return null;
            if(s.length() != 1) throw new RuntimeException("Unable to convert " + s + " to character ! length != 1");
            return s.charAt(0);
        }

        public Type getSource() {
            return String.class;
        }
        public Type getTarget() {
            return Character.class;
        }
    }


    @Extension("PluginSystem@Converter")
    public static class String2Integer implements ExtendedConverter<String, Integer>{
        public Integer convert(String s, ClassLoader loader) {
            return Integer.parseInt(s);
        }
        public Type getSource() {
            return String.class;
        }
        public Type getTarget() {
            return Integer.class;
        }
    }

    @Extension("PluginSystem@Converter")
    public static class String2Long implements ExtendedConverter<String, Long>{
        public Long convert(String s, ClassLoader loader) {
            return Long.parseLong(s);
        }
        public Type getSource() {
            return String.class;
        }
        public Type getTarget() {
            return Long.class;
        }
    }

    @Extension("PluginSystem@Converter")
    public static class String2Byte implements ExtendedConverter<String, Byte>{
        public Byte convert(String s, ClassLoader loader) {
            return Byte.parseByte(s);
        }
        public Type getSource() {
            return String.class;
        }
        public Type getTarget() {
            return Byte.class;
        }
    }

    @Extension("PluginSystem@Converter")
    public static class String2Short implements ExtendedConverter<String, Short>{
        public Short convert(String s, ClassLoader loader) {
            return Short.parseShort(s);
        }
        public Type getSource() {
            return String.class;
        }
        public Type getTarget() {
            return Short.class;
        }
    }

    @Extension("PluginSystem@Converter")
    public static class String2Float implements ExtendedConverter<String, Float>{
        public Float convert(String s, ClassLoader loader) {
            return Float.parseFloat(s);
        }
        public Type getSource() {
            return String.class;
        }
        public Type getTarget() {
            return Float.class;
        }
    }
    @Extension("PluginSystem@Converter")
    public static class String2Double implements ExtendedConverter<String, Double>{
        public Double convert(String s, ClassLoader loader) {
            return Double.parseDouble(s);
        }
        public Type getSource() {
            return String.class;
        }
        public Type getTarget() {
            return Double.class;
        }
    }
}
