/*
 * Pluginsystem
 * Copyright (C) Thasso Griebel
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package org.java.plugin.extension.converter;

import com.google.inject.internal.MoreTypes;
import com.google.inject.util.Types;

import java.lang.reflect.Type;
import java.lang.reflect.WildcardType;
import java.util.ArrayList;
import java.util.Map;
import java.util.HashMap;
import java.io.Serializable;

import org.java.plugin.extension.converter.StringConverter;
import org.java.plugin.extension.annotations.Extension;

/**
 * This converts a String to a type, with respect to generics. The converter need fully qualified classnames!
 * Only a few shortcuts are supported:
 * <pre>
    "List" -> "java.util.List"
    "Map" -> "java.util.Map"
    "Set" -> "java.util.Set"
    "Integer" -> "java.lang.Integer"
    "Float" -> "java.lang.Float"
    "Double" -> "java.lang.Double"
    "Boolean" -> "java.lang.Boolean"
    "Long" -> "java.lang.Long"
    "Number" -> "java.lang.Number"
    "String" -> "java.lang.String"
    "Character" -> "java.lang.Character"
    "Byte" -> "java.lang.Byte"
    "Object" -> "java.lang.Object"         
 * </pre>
 *
 *
 * @author Thasso Griebel (thasso.griebel@gmail.com)
 */
@Extension("PluginSystem@Converter")
public class TypeConverter extends StringConverter<Type>{
    public static final Type[] EMPTY_TYPE_ARRAY = new Type[] {};
    /**
     * shortcuts to classnames
     */
    private static Map<String,String> SHORTCUTS;
    static{
        SHORTCUTS = new HashMap<String, String>();
        SHORTCUTS.put("List", "java.util.List");
        SHORTCUTS.put("Map", "java.util.Map");
        SHORTCUTS.put("Set", "java.util.Set");
        SHORTCUTS.put("Integer", "java.lang.Integer");
        SHORTCUTS.put("Float", "java.lang.Float");
        SHORTCUTS.put("Double", "java.lang.Double");
        SHORTCUTS.put("Boolean", "java.lang.Boolean");
        SHORTCUTS.put("Long", "java.lang.Long");
        SHORTCUTS.put("Number", "java.lang.Number");
        SHORTCUTS.put("String", "java.lang.String");
        SHORTCUTS.put("Character", "java.lang.Character");
        SHORTCUTS.put("Byte", "java.lang.Byte");
        SHORTCUTS.put("Object", "java.lang.Object");        
    }


    /**
     * Converts the given String to a Type using this class' ClassLoader
     *
     * @param value the String to convert
     * @return type representing the String
     */
    public static Type getType(String value){
        return getType(value, TypeConverter.class.getClassLoader());
    }
    /**
     * Converts the given String to a Type using the given ClassLoader
     *
     * @param value the String to convert
     * @param loader the loader used to load classes
     * @return type representing the String
     */
    public static Type getType(String value, ClassLoader loader){
        TypeConverter c = new TypeConverter();
        return c.convert(value,loader);
    }


    public Type convert(String value, ClassLoader loader){
        // XML cehat, translate () to <>
        value = value.replaceAll("\\(","<");
        value = value.replaceAll("\\)",">");


        value = value.trim();
        if(value.lastIndexOf("<") < 0){
            // simple class
            Type c = null;
            if(value.indexOf("[") <0){
                 c = loadClass(value,loader);
            }else{
                c = Types.arrayOf(convert(value.substring(0, value.length()-2),loader));
            }
            return c;
        }else{
            // split < >
            String left = value.substring(0, value.indexOf("<"));
            String right = value.substring(value.indexOf("<")+1, value.lastIndexOf(">"));
            Type ltype = convert(left, loader);
            if(right.indexOf(",") <0){
                Type rtype = convert(right,loader);
                return Types.newParameterizedType(ltype, new Type[]{rtype});
            }else{
                ArrayList<Type> list = new ArrayList<Type>();
                String[] rights = right.split(",");
                for (String s : rights) {
                    list.add(convert(s, loader));
                }
                return Types.newParameterizedType(ltype, list.toArray(new Type[list.size()]));
            }
        }
    }

    private Type loadClass(String value, ClassLoader loader) {
        // check for ? extends or ? super or ?
        if(value.startsWith("?")){
            if(value.indexOf("extends") >= 0){
                String cname = value.split("extends")[1];
                cname = cname.trim();
                if(cname.contains("&")){
                   String[] classes = cname.split("&");
                    ArrayList<Type> cs = new ArrayList<Type>();
                    for (String aClass : classes) {
                        cs.add(loadClass(aClass.trim(), loader));
                    }
                    return new WildcardTypeImpl(cs.toArray(new Type[cs.size()]), null);
                }else{
                    return Types.subtypeOf(loadClass(cname, loader));
                }
            }else if(value.indexOf("super") >= 0){
                String cname = value.split("super")[1];
                cname = cname.trim();
                if(cname.contains("&")){
                    String[] classes = cname.split("&");
                     ArrayList<Type> cs = new ArrayList<Type>();
                     for (String aClass : classes) {
                         cs.add(loadClass(aClass.trim(), loader));
                     }
                     return new WildcardTypeImpl(new Type[]{Object.class}, cs.toArray(new Type[cs.size()]));
                }else{
                    return Types.supertypeOf(loadClass(cname, loader));
                }
            }else{
                return Types.subtypeOf(Object.class);                
            }
        }

        try {
            if(SHORTCUTS.containsKey(value)){
                value = SHORTCUTS.get(value);
            }
            if(loader == null){
                return getClass().getClassLoader().loadClass(value);
            }
            return  loader.loadClass(value);
        } catch (ClassNotFoundException e) {
            throw new RuntimeException("Unable load class '" + value + "'. Make sure you have specified the full class name (including the package)!");
        }
    }

    public Type getTarget() {
        return Type.class;
    }

    /**
     * WildcardType  implementation that supports multiple lower and upper bounds
     */
   public static class WildcardTypeImpl implements WildcardType, Serializable {

    private final Type[] upperBounds;
    private final Type[] lowerBounds;

    public WildcardTypeImpl(Type[] upperBounds, Type[] lowerBounds) {
        this.lowerBounds = lowerBounds;
        this.upperBounds = upperBounds;
    }

    public Type[] getUpperBounds() {
      return this.upperBounds;
    }

    public Type[] getLowerBounds() {
      return lowerBounds != null ? lowerBounds : EMPTY_TYPE_ARRAY;
    }

    @Override public boolean equals(Object other) {
      return other instanceof WildcardType
          && MoreTypes.equals(this, (WildcardType) other);
    }

    @Override public int hashCode() {
      return MoreTypes.hashCode(this);
    }

    @Override public String toString() {
      return MoreTypes.toString(this);
    }

    private static final long serialVersionUID = 0;
  }
}
