/*
 * Pluginsystem
 * Copyright (C) Thasso Griebel
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package org.java.plugin.extension.util;

import com.google.inject.spi.TypeConverter;
import com.google.inject.TypeLiteral;
import org.java.plugin.extension.converter.Converter;
import org.java.plugin.extension.converter.ConverterGraph;

/**
 *
 * Extend this class to provide costom type convertion from within a plugin.
 * The loader will be set to the plugins class loader and you can use it in your
 * convert method.
 *
 * @author Thasso Griebel (thasso.griebel@gmail.com)
 * @see com.google.inject.spi.TypeConverter
 */
public class PluginTypeConverter implements TypeConverter {
    /**
     * The plugins classloader
     */
    private ClassLoader loader;

    private ConverterGraph converters;

    /**
     * Create a new PluginTypeConverter with given loader
     *
     * @param loader the classloader used to resolve resources
     */
    public PluginTypeConverter(ClassLoader loader, ConverterGraph converters){
        super();
        this.loader = loader;
        this.converters = converters;
    }

    /**
     * Convert given String to Object. You can use {@link #getLoader()} to
     * resolve resources from within the current plugin.
     *
     * @param value the string to be converted
     * @param toType the target type 
     * @return object convefrted from given string
     */
    public Object convert(String value, TypeLiteral<?> toType){
        Converter c = converters.getConverter(String.class, toType.getType());
        Object returnValue = null;
        if(c != null){            
             returnValue = c.convert(value, getLoader());
        }
        if(returnValue == null){
            throw new RuntimeException("Converter returned a null value for '"+value+"'. Null values are currently not supported for @Named annotations. Consider using @Nullable instead.");
        }
        return returnValue;
    }

    /**
     * Get the plugin classloader
     *
     * @return classloader the plugin classloader
     */
    protected ClassLoader getLoader() {
        return loader;
    }
}
